/*
  dnsptr -- an example of making DNS PTR queries using libc
  Copyright (C) 2007 Sam Clippinger (samc (at) silence (dot) org)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>

#define MAX_RDNS                        29
#define RDNS_SUFFIX                     ".in-addr.arpa"

#define ERROR_RES_INIT           "ERROR: unable to initialize resolver library."
#define ERROR_DN_EXPAND          "ERROR: dn_expand failed for %s; this could indicate a problem with the nameserver."
#define ERROR_DN_SKIPNAME        "ERROR: dn_skipname failed for %s; this could indicate a problem with the nameserver."
#define ERROR_TYPE_PTR           "ERROR: dns query of %s did not include a PTR record as requested; this could indicate a problem with the name server."
#define ERROR_DNS_RESPONSE       "ERROR: bad or invalid dns response to %s; this could indicate a problem with the name server."

/*
 * The DNS packet format is not well documented outside of the RFCs.  There is
 * almost no sample or tutorial code to be found on the internet outside of the
 * sendmail source code, which is pretty hard to read.
 *
 * Basically, each DNS packet starts with a HEADER structure, defined in
 * arpa/nameser.h (or arpa/nameser_compat.h on Linux).  Most of the time, the
 * header can be skipped.
 *
 * After the header, the nameserver returns all of the "questions" it was asked,
 * so the answers will make sense.  If you're asking more than one "question"
 * per query, this is important.  Otherwise, skip them by finding the size of
 * each question with dn_skipname() and advancing past them.  The number of
 * questions is found in the qdcount field of the header.
 *
 * Next is the answer section, which can contain many answers, though multiple
 * answers may not make much sense for all query types.  The number of answers
 * is found in the ancount field of the header.  Within each answer, the first
 * field is the name that was queried, for reference.  It can be skipped with
 * dn_skipname().
 *
 * After that comes the type in a 16 bit field, then the class in a 16 bit
 * field, then the time-to-live (TTL) in a 32 bit field, then the answer size
 * in a 16 bit field.  The type and size are important; the class and the ttl
 * can usually be ignored.  The format of the rest of the answer field is
 * different depending on the type.
 *
 * IF THE TYPE IS T_TXT:
 * The first 8 bits are an unsigned integer indicating the total length of
 * the text response.  The following bytes are the ASCII text of the response.
 *
 * IF THE TYPE IS T_PTR:
 * All of the bytes (up to the length given by size) are the name of the PTR
 * entry but the subdomains are separated by 8 bit unsigned integer values
 * indicating the length of each subdomain portion.  For example, the name
 * foo.barney.example.com would be encoded by an ASCII 3, the characters
 * "foo", an ASCII 6, the characters "barney", an ASCII 7, the characters
 * "example", an ASCII 3 and the characters "com".  The size field would be
 * 23.
 *
 * IF THE TYPE IS T_CNAME:
 * All of the bytes are the compressed name of the CNAME entry.  They can be
 * decoded with dn_expand().
 *
 * IF THE TYPE IS T_MX:
 * Each answer begins with an unsigned 16 bit integer indicating the preference
 * of the mail server (lower preferences should be contacted first).  The
 * remainder of the answer is the mail server name.  It can be decoded with
 * dn_expand().
 */
int dns_ptr_lookup(char *target_name)
  {
  static int initialized = 0;
  int return_value;
  int i;
  unsigned char answer[PACKETSZ];
  unsigned char host[128];
  unsigned char *answer_ptr;
  unsigned char *tmp_ptr;
  int answer_length;
  int size;
  int txt_length;
  int type;
  int num_questions;
  int num_answers;

  return_value = 0;

  if (!initialized)
    {
    if (res_init() >= 0)
      initialized = 1;
    else
      printf(ERROR_RES_INIT);
    }

  if (initialized)
    {
    memset(answer, 0, PACKETSZ);

    if ((target_name != NULL) &&
        ((answer_length = res_search(target_name, C_IN, T_PTR, answer, PACKETSZ)) >= 0))
      {
      answer_ptr = answer + sizeof(HEADER);

      // Skip the questions
      num_questions = ntohs((unsigned short)((HEADER *)&answer)->qdcount);
      for (i = 0; i < num_questions; i++)
        if ((size = dn_skipname(answer_ptr, answer + answer_length)) >= 0)
          answer_ptr += size + QFIXEDSZ;
        else
          break;

      if (i == num_questions)
        {
        num_answers = ntohs((unsigned short)((HEADER *)&answer)->ancount);
        for (i = 0; (i < num_answers) && !return_value; i++)
          if ((size = dn_skipname(answer_ptr, answer + answer_length)) >= 0)
            {
            answer_ptr += size;
            GETSHORT(type, answer_ptr);
            answer_ptr += INT16SZ; // class
            answer_ptr += INT32SZ; // ttl
            GETSHORT(size, answer_ptr);

            if (type == T_PTR)
              {
              if (size > 2)
                {
                tmp_ptr = answer_ptr;
                while (((tmp_ptr - answer_ptr) < size) &&
                       ((txt_length = tmp_ptr[0]) > 0))
                  {
                  tmp_ptr[0] = '.';
                  tmp_ptr += txt_length + 1;
                  }

                printf("%.*s\n", size - 2, answer_ptr + 1);

                return_value = 1;
                answer_ptr += size;
                }
              else
                {
                printf(ERROR_DNS_RESPONSE, target_name);
                break;
                }
              }
            else if (type == T_CNAME)
              {
              if ((size = dn_expand(answer, answer + answer_length, answer_ptr, (char *)host, sizeof(host))) >= 0)
                {
                return_value = dns_ptr_lookup((char *)host);
                answer_ptr += size;
                }
              else
                {
                printf(ERROR_DN_EXPAND, target_name);
                break;
                }
              }
            else
              {
              printf(ERROR_TYPE_PTR, target_name);
              break;
              }
            }
          else
            {
            printf(ERROR_DN_SKIPNAME, target_name);
            break;
            }
        }
      else
        printf(ERROR_DN_SKIPNAME, target_name);
      }
    }

  return(return_value);
  }

int dns_ptr(char *target_ip)
  {
  int return_value;
  char ip_octets[4][4];
  char rdns_name[MAX_RDNS + 1];

  return_value = 0;

  if ((target_ip != NULL) &&
      (sscanf(target_ip, "%3[0-9].%3[0-9].%3[0-9].%3[0-9]", ip_octets[0], ip_octets[1], ip_octets[2], ip_octets[3]) == 4) &&
      snprintf(rdns_name, MAX_RDNS, "%s.%s.%s.%s%s", ip_octets[3], ip_octets[2], ip_octets[1], ip_octets[0], RDNS_SUFFIX))
    return_value = dns_ptr_lookup(rdns_name);

  return(return_value);
  }

int main(int argc, char *argv[])
  {
  int i;

  if (argc > 1)
    for (i = 1; i < argc; i++)
      dns_ptr(argv[i]);
  else
    printf("USAGE: dns_ptr IP\n");

  return(0);
  }
